import logging
import time

from fido2.client import U2fClient, ClientError
from fido2.hid import CtapHidDevice

from nd_okta_auth.factor import Factor, FactorVerificationFailed

log = logging.getLogger(__name__)


class U2fFactor(Factor):
    def name(self):
        return "u2f"

    def _get_devices(self):
        return CtapHidDevice.list_devices()

    def _get_client(self, dev, appId):
        return U2fClient(dev, appId)

    def verify(self, fid, state_token, sleep):
        """Validates user with Okta using user's U2F hardware device.

        This method is meant to be called by self.auth() if a Login session
        requires MFA, and the users profile supports U2F.

        We wait for a U2F device to be plugged into USB, request a challenge
        nonce from Okta, pass the challenge to the U2F device, and finally
        send back the challenge response to Okta.  If its accepted, we write
        out our SessionToken.

        Args:
            fid: Okta Factor ID used to trigger the push
            state_token: State Token allowing us to trigger the push
            sleep: amount of seconds to wait between checking for U2F hardware
                   keep to be plugged in.
        """
        path = self.verify_path.format(fid=fid)

        # Wait for U2F device to be plugged in before continuing on
        dev = None
        while True:
            dev = next(self._get_devices(), None)
            if dev:
                break
            log.info("Waiting for FIDO U2F device to be plugged in.")
            time.sleep(sleep)

        # Request U2F nonce/challenge from Okta
        log.info("Requesting U2F challenge nonce from Okta")
        data = {"fid": fid, "stateToken": state_token}
        ret = self._request(path, data)

        if ret["status"] != "MFA_CHALLENGE":
            raise FactorVerificationFailed("Expected MFA challenge")

        # Get U2F device to sign nonce/challenge
        appId = self.base_url
        client = self._get_client(dev, appId)

        nonce = ret["_embedded"]["factor"]["_embedded"]["challenge"]["nonce"]
        credentialId = ret["_embedded"]["factor"]["profile"]["credentialId"]
        registered_keys = [{"version": "U2F_V2", "keyHandle": credentialId}]

        log.warning("Touch your authenticator device now...")

        try:
            r = client.sign(appId, nonce, registered_keys)
        except ClientError:
            raise FactorVerificationFailed(
                "U2F devices failed to "
                "sign request. Have you "
                'registered it with "{}"?'.format(appId)
            )

        # Send challenge response back to Okta
        data = {
            "stateToken": state_token,
            "clientData": r.get("clientData"),
            "signatureData": r.get("signatureData"),
        }

        ret = self._request(path, data)

        if ret.get("status") != "SUCCESS":
            raise FactorVerificationFailed()

        return ret



"""
okta
^^^^

Handles the initial Okta authentication - throws appropriate errors in the
events of bad passwords, MFA requirements, etc.
"""

from __future__ import unicode_literals

import base64
import logging
import sys

import bs4
import requests

from nd_okta_auth import factor, base_client

if sys.version_info[0] < 3:  # Python 2
    from exceptions import Exception

log = logging.getLogger(__name__)


class UnknownError(Exception):
    """Some Expected Return Was Received"""


class EmptyInput(base_client.BaseException):
    """Invalid Input - Empty String Detected"""


class InvalidPassword(base_client.BaseException):
    """Invalid Password"""


class ExhaustedFactors(base_client.BaseException):
    """Failed to authenticate user with any factor"""


class Okta(base_client.BaseOktaClient):
    """Base Okta Login Object with MFA handling.

    This base login object handles connecting to Okta, authenticating a user,
    and optionally triggering MFA Authentication. No application specific logic
    is here, just the initial authentication and creation of a
    cookie-authenticated requests.Session() object.

    See OktaSaml for a more useful object.
    """

    def __init__(self, organization, username, password):
        base_client.BaseOktaClient.__init__(self, organization)
        log.debug("Base URL Set to: {url}".format(url=self.base_url))

        # Validate the inputs are reasonably sane
        for input in (organization, username, password):
            if input == "" or input is None:
                raise EmptyInput()

        self.username = username
        self.password = password
        self.supported_factors = factor.factors(organization)
        self.session_token = None

    def set_token(self, ret):
        """Parses an authentication response and stores the token.

        Parses a SUCCESSFUL authentication response from Okta and stores the
        token.

        args:
            ret: The response from Okta that we know is successful and contains
            a sessionToken
        """
        firstName = ret["_embedded"]["user"]["profile"]["firstName"]
        lastName = ret["_embedded"]["user"]["profile"]["lastName"]
        log.info(
            "Successfully authed {firstName} {lastName}".format(
                firstName=firstName, lastName=lastName
            )
        )
        self.session_token = ret["sessionToken"]

    def auth(self):
        """Performs an initial authentication against Okta.

        The initial Okta Login authentication is handled here - and optionally
        MFA authentication is triggered. If successful, this method stores a
        SessionToken. This SessionToken can be used to initiate a call to the
        "Embed Link" of an Okta Application.

        **Note ... Undocumented/Unclear Okta Behavior**
        If you use the SessionToken only to make your subsequent requests, its
        usable only once and then it expires. However, if you combine it with a
        long-lived SID cookie (which we do, by using reqests.Session() to make
        all of our web requests), then that SessionToken can be redeemd many
        times as long as you do it through the "Embed Links". See the OktaSaml
        client for an example.

            https://developer.okta.com/use_cases/authentication/
            session_cookie#visit-an-embed-link-with-the-session-token
        """
        path = "/authn"
        data = {"username": self.username, "password": self.password}
        try:
            ret = self._request(path, data)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 401:
                raise InvalidPassword()

        status = ret.get("status", None)

        if status == "SUCCESS":
            self.set_token(ret)
            return

        if status == "MFA_ENROLL" or status == "MFA_ENROLL_ACTIVATE":
            log.warning("User {u} needs to enroll in 2FA first".format(u=self.username))
            raise UnknownError()

        if status == "MFA_REQUIRED" or status == "MFA_CHALLENGE":
            # Factors enabled by the user
            enabled_factors = ret["_embedded"]["factors"]

            # Loop through locally supported factors
            for supported_factor in self.supported_factors:
                # filter enabled factors against support factors
                filtered_factors = list(
                    filter(
                        lambda x: x.get("factorType") == supported_factor.name(),
                        enabled_factors,
                    )
                )

                # Try authenticating with each enabled factor
                for enabled_factor in filtered_factors:
                    log.info(
                        "Authenticating with factor: {} id {}".format(
                            supported_factor.name(), enabled_factor["id"]
                        )
                    )

                    try:
                        ret = supported_factor.verify(
                            enabled_factor["id"], ret["stateToken"], sleep=1
                        )
                        self.set_token(ret)
                        return
                    except KeyboardInterrupt:
                        # Allow users to use MFA Push by breaking
                        # out of waiting for U2F device.
                        log.info(
                            "User skipping factor: {}".format(supported_factor.name())
                        )
                        continue
                    except factor.FactorVerificationFailed as e:
                        # Non fatal error that a factor failed to
                        # be verified.
                        log.error(e)
                        continue
                    except requests.exceptions.ReadTimeout:
                        log.error(
                            "HTTP timeout contacting Okta at {}".format(self.base_url)
                        )
                        continue

        raise ExhaustedFactors("Failed to verify with any MFA factor")


class OktaSaml(Okta):
    def assertion(self, saml):
        assertion = ""
        soup = bs4.BeautifulSoup(saml, "html.parser")
        for inputtag in soup.find_all("input"):
            if inputtag.get("name") == "SAMLResponse":
                assertion = inputtag.get("value")
        return base64.b64decode(assertion)

    def get_assertion(self, appid, apptype):
        path = "{url}/home/{apptype}/{appid}".format(
            url=self.base_url, apptype=apptype, appid=appid
        )
        resp = self.session.get(path, params={"onetimetoken": self.session_token})
        log.debug(resp.__dict__)

        try:
            resp.raise_for_status()
        except requests.exceptions.HTTPError as e:
            log.error("Unknown error: {msg}".format(msg=str(e.response.__dict__)))
            raise UnknownError()

        return self.assertion(resp.text)



# -*- coding: utf-8 -*-
#
# Credits: This code was copied from
# https://github.com/ThoughtWorksInc/aws_role_credentials
#
# Copyright (c) 2015, Peter Gillard-Moss
# All rights reserved.

# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.

# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import base64
import xml.etree.ElementTree as ET


class SamlAssertion:
    def __init__(self, assertion):
        self.assertion = assertion

    @staticmethod
    def split_roles(roles):
        return [(y.strip()) for y in roles.text.split(",")]

    @staticmethod
    def sort_roles(roles):
        return sorted(roles, key=lambda role: "saml-provider" in role)

    def roles(self):
        attributes = ET.fromstring(self.assertion).iter(
            "{urn:oasis:names:tc:SAML:2.0:assertion}Attribute"
        )

        name = "https://aws.amazon.com/SAML/Attributes/Role"
        roles_attributes = [x for x in attributes if x.get("Name") == name]

        roles_values = [
            (x.iter("{urn:oasis:names:tc:SAML:2.0:assertion}AttributeValue"))
            for x in roles_attributes
        ]

        return [
            (dict(zip(["role", "principle"], self.sort_roles(self.split_roles(x)))))
            for x in roles_values[0]
        ]

    def encode(self):
        return base64.b64encode(self.assertion).decode()



import abc
import logging

from nd_okta_auth import base_client


class FactorVerificationFailed(base_client.BaseException):
    """Failed to authenticate with second factor"""


log = logging.getLogger(__name__)


class Factor(base_client.BaseOktaClient):
    __metaclass__ = abc.ABCMeta

    verify_path = "/authn/factors/{fid}/verify"

    def __init__(self, organization):
        base_client.BaseOktaClient.__init__(self, organization)

    @abc.abstractmethod
    def name(self):
        """Name of the second factor. Must be same as Okta's `factorType`"""
        return

    @abc.abstractmethod
    def verify(self, fid, state_token, sleep):
        """Verify a user with a second factor."""
        return


def factors(organization):
    from nd_okta_auth.factor.u2f import U2fFactor  # noqa: F401
    from nd_okta_auth.factor.push import PushFactor  # noqa: F401
    from nd_okta_auth.factor.totp import TotpFactor  # noqa: F401
    from nd_okta_auth.factor.webauthn import WebauthnFactor  # noqa: F401

    return [
        WebauthnFactor(organization),
        U2fFactor(organization),
        PushFactor(organization),
        TotpFactor(organization),
    ]



import logging
import requests
import sys

if sys.version_info[0] < 3:  # Python 2
    from exceptions import Exception

log = logging.getLogger(__name__)

BASE_URL = "https://{organization}.okta.com"


class BaseException(Exception):
    """Base Exception for Okta Auth"""


class BaseOktaClient(object):
    def __init__(self, organization):
        self.base_url = BASE_URL.format(organization=organization)
        self.session = requests.Session()

    def _request(self, path, data=None):
        """Basic URL Fetcher for Okta

        Any HTTPError is raised immediately, otherwise the response is parsed
        as JSON and passed back as a dictionary.

        Args:
            path: The path at the base url to call
            data: Optional data to pass in as Post parameters

        Returns:
            The response in dict form.
        """
        headers = {"Accept": "application/json", "Content-Type": "application/json"}

        if path.startswith("http"):
            url = path
        else:
            url = "{base}/api/v1{path}".format(base=self.base_url, path=path)

        resp = self.session.post(
            url=url, headers=headers, json=data, allow_redirects=False
        )

        resp_obj = resp.json()
        log.debug(resp_obj)

        resp.raise_for_status()
        return resp_obj



import base64
import logging
import time

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

from fido2.client import Fido2Client, ClientError
from fido2.hid import CtapHidDevice
from fido2.webauthn import PublicKeyCredentialRequestOptions

from fido2.utils import websafe_decode

from nd_okta_auth.factor import Factor, FactorVerificationFailed

log = logging.getLogger(__name__)


class WebauthnFactor(Factor):
    def name(self):
        return "webauthn"

    def _get_devices(self):
        return CtapHidDevice.list_devices()

    def _get_client(self, dev, appId):
        return Fido2Client(dev, appId)

    def verify(self, fid, state_token, sleep):
        """Validates user with Okta using user's webauthn hardware device.

        This method is meant to be called by self.auth() if a Login session
        requires MFA, and the users profile supports webauthn.

        We wait for a webauthn device to be plugged into USB, request a
        challenge nonce from Okta, pass the challenge to the webauthn device,
        and finally send back the challenge response to Okta.  If its accepted,
        we write out our SessionToken.

        Args:
            fid: Okta Factor ID used to trigger the push
            state_token: State Token allowing us to trigger the push
            sleep: amount of seconds to wait between checking for webauthn
                   hardware keep to be plugged in.
        """
        path = self.verify_path.format(fid=fid)

        # Wait for webauthn device to be plugged in before continuing on
        dev = None
        while True:
            dev = next(self._get_devices(), None)
            if dev:
                break
            log.info("Waiting for FIDO webauthn device to be plugged in.")
            time.sleep(sleep)

        # Request webauthn nonce/challenge from Okta
        log.info("Requesting webauthn challenge nonce from Okta")
        data = {"fid": fid, "stateToken": state_token}
        ret = self._request(path, data)

        if ret["status"] != "MFA_CHALLENGE":
            raise FactorVerificationFailed("Expected MFA challenge")

        # Get webauthn device to sign nonce/challenge
        appId = self.base_url
        client = self._get_client(dev, appId)
        nonce = ret["_embedded"]["factor"]["_embedded"]["challenge"]["challenge"]
        credential_id = ret["_embedded"]["factor"]["profile"]["credentialId"]
        # Add extra padding to credential_id as it may not have enough
        credential_id = base64.urlsafe_b64decode(credential_id + "====")

        allow_list = [{"type": "public-key", "id": credential_id}]
        rp_id = urlparse(appId).hostname

        log.warning("Touch your authenticator device now...")
        try:
            challenge = websafe_decode(nonce)
            options = PublicKeyCredentialRequestOptions(
                challenge=challenge, rp_id=rp_id, allow_credentials=allow_list
            )

            assertions, client_data = client.get_assertion(options)
        except ClientError:
            raise FactorVerificationFailed(
                "webauthn devices failed to "
                "sign request. Have you "
                'registered it with "{}"?'.format(appId)
            )

        # Send challenge response back to Okta
        assert len(assertions) == 1
        ad = base64.b64encode(assertions[0].auth_data)
        cd = base64.b64encode(client_data)
        sig = base64.b64encode(assertions[0].signature)

        data = {
            "stateToken": state_token,
            "clientData": cd.decode(),
            "authenticatorData": ad.decode(),
            "signatureData": sig.decode(),
        }

        ret = self._request(path, data)

        if ret.get("status") != "SUCCESS":
            raise FactorVerificationFailed()

        return ret



from __future__ import unicode_literals
import getpass
import logging
import sys
import time
import requests
from builtins import input

import rainbow_logging_handler

from nd_okta_auth import okta
from nd_okta_auth import aws
from nd_okta_auth.metadata import __desc__, __version__


def user_input(text):
    """Wraps input() making testing support of py2 and py3 easier"""
    return input(text)


def setup_logging():
    """Returns back a pretty color-coded logger"""
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    handler = rainbow_logging_handler.RainbowLoggingHandler(sys.stdout)
    fmt = "%(asctime)-10s (%(levelname)s) %(message)s"
    formatter = logging.Formatter(fmt)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


def login(
    aws_profile: str,
    okta_appid: str,
    okta_org: str,
    username: str,
    reup: bool,
    debug: bool = False,
):
    # Generate our logger first, and write out our app name and version
    log = setup_logging()
    log.info("%s v%s" % (__desc__, __version__))

    if debug:
        log.setLevel(logging.DEBUG)

    # Ask the user for their password.. we do this once at the beginning, and
    # we keep it in memory for as long as this tool is running. Its never ever
    # written out or cached to disk anywhere.
    password = getpass.getpass()

    # Generate our initial OktaSaml client and handle any exceptions thrown.
    # Generally these are input validation issues.
    try:
        okta_client = okta.OktaSaml(okta_org, username, password)
    except okta.EmptyInput:
        log.error("Cannot enter a blank string for any input")
        raise

    # Authenticate the Okta client. If necessary, we will ask for MFA input.
    try:
        okta_client.auth()
    except okta.InvalidPassword:
        log.error("Invalid Username ({user}) or Password".format(user=username))
        raise
    except okta.ExhaustedFactors as e:
        log.error(e)
        raise

    # Once we're authenticated with an OktaSaml client object, we can use that
    # object to get a fresh SAMLResponse repeatedly and refresh our AWS
    # Credentials.
    session = None
    role_selection = None
    while True:
        # If an AWS Session object has been created already, lets check if its
        # still valid. If it is, sleep a bit and skip to the next execution of
        # the loop.
        if session and session.is_valid:
            log.debug("Credentials are still valid, sleeping")
            time.sleep(15)
            continue

        log.info("Getting SAML Assertion from {org}".format(org=okta_org))

        try:
            assertion = okta_client.get_assertion(
                appid=okta_appid, apptype="amazon_aws"
            )
            session = aws.Session(assertion, profile=aws_profile)

            # If role_selection is set we're in a reup loop. Re-set the role on
            # the session to prevent the user being prompted for the role again
            # on each subsequent renewal.
            if role_selection is not None:
                session.set_role(role_selection)

            session.assume_role()

        except aws.MultipleRoles:
            log.warning("Multiple AWS roles found; please select one")
            roles = session.available_roles()
            for role_index, role in enumerate(roles):
                print("[{}] Role: {}".format(role_index, role["role"]))
            role_selection = user_input("Select a role from above: ")
            session.set_role(role_selection)
            session.assume_role()
        except requests.exceptions.ConnectionError:
            log.warning("Connection error... will retry")
            time.sleep(5)
            continue

        # If we're not running in re-up mode, once we have the assertion
        # and creds, go ahead and quit.
        if not reup:
            break

        log.info("Reup enabled, sleeping...")
        time.sleep(5)



import logging
import time

from nd_okta_auth.factor import Factor, FactorVerificationFailed

log = logging.getLogger(__name__)


class PushFactor(Factor):
    def name(self):
        return "push"

    def verify(self, fid, state_token, sleep):
        """Triggers an Okta Push Verification and waits.

        This method is meant to be called by self.auth() if a Login session
        requires MFA, and the users profile supports Okta Push with Verify.

        We trigger the push, and then immediately go into a wait loop. Each
        time we loop around, we pull the latest status for that push event. If
        its Declined, we will throw an error. If its accepted, we write out our
        SessionToken.

        Args:
            fid: Okta Factor ID used to trigger the push
            state_token: State Token allowing us to trigger the push
            sleep: amount of time to sleep between checking for push status
        """
        log.warning("Okta Verify Push being sent...")
        path = self.verify_path.format(fid=fid)
        data = {"fid": fid, "stateToken": state_token}
        ret = self._request(path, data)

        while ret["status"] != "SUCCESS":
            log.info("Waiting for Okta Verification...")
            time.sleep(sleep)

            if ret.get("factorResult", "REJECTED") == "REJECTED":
                raise FactorVerificationFailed("Okta Verify Push REJECTED")

            links = ret.get("_links")
            ret = self._request(links["next"]["href"], data)

        return ret



import logging

import requests
from six.moves import input

from nd_okta_auth.factor import Factor

log = logging.getLogger(__name__)


def user_input(text):
    """Wraps input() making testing support of py2 and py3 easier"""
    return input(text)


class TotpFactor(Factor):
    def get_passcode(self):
        return user_input("Time-based one-time passcode: ")

    def name(self):
        return "token:software:totp"

    def verify(self, fid, state_token, sleep):
        """Validates an Okta user with Passcode-based MFA.

        Takes in the supplied Factor ID (fid), State Token and user supplied
        Passcode, and validates the auth. If successful, sets the session
        token. If invalid, raises an exception.

        Args:
            fid: Okta Factor ID (returned in the PasscodeRequired exception)
            state_token: State Tken (returned in the PasscodeRequired
            exception)
            sleep: not used
        Returns:
            Response from okta
        """

        while True:
            passcode = self.get_passcode()
            if len(passcode) != 6:
                log.error("Passcodes must be 6 digits")
                continue

            path = "/authn/factors/{fid}/verify".format(fid=fid)
            data = {"fid": fid, "stateToken": state_token, "passCode": passcode}
            try:
                return self._request(path, data)
            except requests.exceptions.HTTPError as e:
                if e.response.status_code == 403:
                    log.error("Invalid Passcode Detected")
                    continue
                raise e



# Copyright 2017 Nextdoor.com, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

from setuptools import Command
from setuptools import setup
from setuptools import find_packages

from nd_okta_auth.metadata import __desc__, __version__

PACKAGE = 'nd_okta_auth'
DIR = os.path.dirname(os.path.realpath(__file__))


setup(
    name=PACKAGE,
    version=__version__,
    description=__desc__,
    long_description=open('%s/README.md' % DIR).read(),
    author='Nextdoor Engineering',
    author_email='eng@nextdoor.com',
    url='https://github.com/Nextdoor/nd_okta_auth',
    download_url='http://pypi.python.org/pypi/%s#downloads' % PACKAGE,
    license='Apache License, Version 2.0',
    keywords='apache',
    packages=find_packages(),
    test_suite='nose.collector',
    tests_require=open('%s/requirements.test.txt' % DIR).readlines(),
    setup_requires=[],
    install_requires=open('%s/requirements.txt' % DIR).readlines(),
    entry_points={
        'console_scripts': [
            'nd_okta_auth = nd_okta_auth.main:entry_point'
        ],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Software Development',
        'License :: OSI Approved :: Apache Software License',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Operating System :: POSIX',
        'Natural Language :: English',
    ]
)



#!/usr/bin/env python

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2017 Nextdoor.com, Inc
import argparse
from future.moves import sys
from nd_okta_auth import auth
from nd_okta_auth.metadata import __version__


def get_config_parser(argv):
    """Returns a configured ArgumentParser for the CLI options"""
    epilog = (
        "**Application ID**\n"
        "The ApplicationID is actually a two part piece of the redirect URL \n"
        "that Okta uses when you are logged into the Web UI. If you mouse \n"
        "over the appropriate Application and see a URL that looks like \n"
        "this. \n"
        "\n"
        "\thttps://foobar.okta.com/home/amazon_aws/0oaciCSo1d8/123?...\n"
        "\n"
        'You would enter in "0oaciCSo1d8/123" as your Application ID.\n'
    )

    arg_parser = argparse.ArgumentParser(
        prog=argv[0],
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=epilog,
        description="Okta Auther",
    )

    # Get rid of the default optional arguments section that always shows up.
    # Its not necessary, and confusing to have optional arguments listed first.
    #   https://stackoverflow.com/questions/24180527/
    #   argparse-required-arguments-listed-under-optional-arguments
    arg_parser._action_groups.pop()

    required_args = arg_parser.add_argument_group("required arguments")
    required_args.add_argument(
        "-o",
        "--org",
        type=str,
        help=(
            "Okta Organization Name - ie, if your login"
            " URL is https://foobar.okta.com, enter in "
            "foobar here"
        ),
        required=True,
    )
    required_args.add_argument(
        "-u",
        "--username",
        type=str,
        help=(
            "Okta Login Name - either bob@foobar.com, "
            "or just bob works too, depending on your "
            "organization settings."
        ),
        required=True,
    )
    required_args.add_argument(
        "-a",
        "--appid",
        type=str,
        help=(
            'The "redirect link" Application ID  - '
            "this can be found by mousing over the "
            "application in Okta's Web UI. See "
            "details below for more help."
        ),
        required=True,
    )

    optional_args = arg_parser.add_argument_group("optional arguments")
    optional_args.add_argument("-V", "--version", action="version", version=__version__)
    optional_args.add_argument(
        "-D",
        "--debug",
        action="store_true",
        help=(
            "Enable DEBUG logging - note, this is "
            "extremely verbose and exposes credentials "
            "so be careful here!"
        ),
        default=False,
    )
    optional_args.add_argument(
        "-r",
        "--reup",
        action="store_true",
        help=("Automatically re-up the AWS creds before" "they expire."),
        default=0,
    )
    optional_args.add_argument(
        "-n", "--name", type=str, help="AWS Profile Name", default="default"
    )

    config = arg_parser.parse_args(args=argv[1:])
    return config


def entry_point():
    """Zero-argument entry point for use with setuptools/distribute."""
    config = get_config_parser(sys.argv)
    raise SystemExit(
        auth.login(
            aws_profile=config.name,
            okta_appid=config.appid,
            okta_org=config.org,
            username=config.username,
            reup=config.reup,
            debug=config.debug,
        )
    )


if __name__ == "__main__":
    entry_point()



"""
aws
^^^

Simple module for writing generating and writing out AWS Credentials into your
~/.aws/credentials file with a supplied Saml assertion.

Credits: This code base was almost entirely stolen from
https://github.com/ThoughtWorksInc/aws_role_credentials. It continues to be
modified from the original code, but thanks a ton to the original writers at
Thought Works Inc.
"""

from __future__ import unicode_literals
from builtins import str
import configparser
import datetime
import logging
import os
from os.path import expanduser
import xml

import boto3
from nd_okta_auth.aws_saml import SamlAssertion

log = logging.getLogger(__name__)


class BaseException(Exception):
    """Base AWS SAML Exception"""


class InvalidSaml(BaseException):
    """Raised when the SAML Assertion is invalid for some reason"""


class MultipleRoles(BaseException):
    """Raised when AWS offers multiple roles"""


class Credentials(object):

    """Simple AWS Credentials Profile representation.

    This object reads in an Amazon ~/.aws/credentials file, and then allows you
    to write out credentials into different Profile sections.
    """

    def __init__(self, filename):
        self.filename = filename

    def _add_profile(self, name, profile):
        config = configparser.ConfigParser(interpolation=None)
        try:
            config.read_file(open(self.filename, "r"))
        except IOError:
            pass

        if not config.has_section(name):
            config.add_section(name)

        [(config.set(name, k, v)) for k, v in profile.items()]
        with open(self.filename, "w+") as configfile:
            os.chmod(self.filename, 0o600)
            config.write(configfile)

    def add_profile(self, name, region, access_key, secret_key, session_token):
        """Writes out a set of AWS Credentials to disk.

        args:
            name: The profile name to write to
            region: The region to use as the default region for this profile
            access_key: The AWS_ACCESS_KEY_ID
            secret_key: The AWS_SECRET_ACCESS_KEY
            session_token: The AWS_SESSION_TOKEN
        """
        name = str(name)
        self._add_profile(
            name,
            {
                "output": "json",
                "region": str(region),
                "aws_access_key_id": str(access_key),
                "aws_secret_access_key": str(secret_key),
                "aws_security_token": str(session_token),
                "aws_session_token": str(session_token),
            },
        )

        log.info(
            'Wrote profile "{name}" to {file}'.format(name=name, file=self.filename)
        )


class Session(object):

    """Amazon Federated Session Generator.

    This class is used to contact Amazon with a specific SAML Assertion and
    get back a set of temporary Federated credentials. These credentials are
    written to disk (using the Credentials object above).

    This object is meant to be used once -- as SAML Assertions are one-time-use
    objects.
    """

    def __init__(
        self, assertion, credential_path="~/.aws", profile="default", region="us-east-1"
    ):
        cred_dir = expanduser(credential_path)
        cred_file = os.path.join(cred_dir, "credentials")

        boto_logger = logging.getLogger("botocore")
        boto_logger.setLevel(logging.WARNING)

        if not os.path.exists(cred_dir):
            log.info("Creating missing AWS Credentials dir {dir}".format(dir=cred_dir))
            os.makedirs(cred_dir)

        self.sts = boto3.client("sts")

        self.profile = profile
        self.region = region

        self.assertion = SamlAssertion(assertion)
        self.writer = Credentials(cred_file)

        # Populated by self.assume_role()
        self.aws_access_key_id = None
        self.aws_secret_access_key = None
        self.aws_session_token = None
        self.expiration = None
        self.session_token = None
        self.role = None

    @property
    def is_valid(self):
        """Returns True if the Session is still valid.

        Takes the current time (in UTC) and compares it to the Expiration time
        returned by Amazon. Adds a 10 minute buffer to make sure that we start
        working to renew the creds far before they really expire and break.

        Args:
            now: A datetime.datetime() object (likely
            datetime.datetime.utcnow())
            buffer: Number of seconds before the actual expiration before we
            start returning false.

        Returns:
            Bool
        """
        # Consider the tokens expired when they have 10m left
        try:
            buffer = datetime.timedelta(seconds=600)
            now = datetime.datetime.utcnow()
            expir = datetime.datetime.strptime(
                str(self.expiration), "%Y-%m-%d %H:%M:%S+00:00"
            )

            return (now + buffer) < expir
        except ValueError:
            return False

    def set_role(self, role_index):
        """Sets the role based on the supplied index value"""
        self.role = self.assertion.roles()[int(role_index)]

    def available_roles(self):
        """Returns the roles availble from AWS"""
        return self.assertion.roles()

    def assume_role(self):
        """Use the SAML Assertion to actually get the credentials.

        Uses the supplied (one time use!) SAML Assertion to go out to Amazon
        and get back a set of temporary credentials. These are written out to
        disk and can be used for an hour before they need to be replaced.
        """
        if self.role is None:
            try:
                if len(self.assertion.roles()) > 1:
                    raise MultipleRoles
                self.role = self.assertion.roles()[0]
            except xml.etree.ElementTree.ParseError:
                log.error("Could not find any Role in the SAML assertion")
                log.error(self.assertion.__dict__)
                raise InvalidSaml()

        log.info("Assuming role: {}".format(self.role["role"]))

        session = self.sts.assume_role_with_saml(
            RoleArn=self.role["role"],
            PrincipalArn=self.role["principle"],
            SAMLAssertion=self.assertion.encode(),
        )
        creds = session["Credentials"]

        self.aws_access_key_id = creds["AccessKeyId"]
        self.aws_secret_access_key = creds["SecretAccessKey"]
        self.session_token = creds["SessionToken"]
        self.expiration = creds["Expiration"]

        self._write()

    def _write(self):
        """Writes out our secrets to the Credentials object"""
        self.writer.add_profile(
            name=self.profile,
            region=self.region,
            access_key=self.aws_access_key_id,
            secret_key=self.aws_secret_access_key,
            session_token=self.session_token,
        )
        log.info("Session expires at {time}".format(time=self.expiration))
